/* 
 * $Id: beetooclient.cpp,v 1.6 2003/06/14 17:28:58 mikelima Exp $
 *
 * BeeToo KWin Client
 * based on the B-II KWin Client
 *
 * Changes:
 *   Customizable button positions by Karol Szwed <gallium@kde.org>
 *   Thin frame in fixed size windows, customizable menu double click action
 *   and title bar gradient support by Luciano Montanaro <mikelima@virgilio.it>
 */ 

#include "beetooclient.h"
#include <qlayout.h>
#include <qstyle.h>
#include <qdrawutil.h>
#include <kpixmapeffect.h>
#include <kimageeffect.h>
#include <kdrawutil.h>
#include <klocale.h>
#include <kconfig.h>
#include <qimage.h>
#include <qbitmap.h>
#include <kstandarddirs.h>
#include <kicontheme.h>
#include <kiconeffect.h>
#include <kwin/workspace.h>
#include <kwin/options.h>

using namespace KWinInternal;

namespace B2 {

#include "bitmaps.h"

enum { 
    Norm = 0, 
    Hover, Down, INorm, IHover, IDown, 
    NumStates 
};

enum {
    P_CLOSE = 0, 
    P_MAX, P_NORMALIZE, P_ICONIFY, P_PINUP, P_MENU, P_HELP, 
    P_NUM_BUTTON_TYPES
};

#define NUM_PIXMAPS ((P_HELP + 1) * NumStates)

static KPixmap *pixmap[NUM_PIXMAPS];

enum {
    TitleActive = 0,
    TitleInactive,
    LeftActive,
    LeftInactive,
    RightActive,
    RightInactive,
    NumTitlePixmaps
};

static QPixmap* titleGradient[NumTitlePixmaps];
static int endWidth[2] = { 3, 3 };

//active
#define PIXMAP_A(i)  (pixmap[(i) * NumStates + Norm])
//active, hover
#define PIXMAP_AH(i) (pixmap[(i) * NumStates + Hover])
//active, down
#define PIXMAP_AD(i) (pixmap[(i) * NumStates + Down])
//inactive
#define PIXMAP_I(i)  (pixmap[(i) * NumStates + INorm])
//inactive, hover
#define PIXMAP_IH(i) (pixmap[(i) * NumStates + IHover])
//inactive, down
#define PIXMAP_ID(i) (pixmap[(i) * NumStates + IDown])

static QString tileName;

static int thickness = 3;
static int menu_dbl_click_op = 0;

static bool pixmaps_created = false;
static bool colored_frame = false;
static bool do_draw_handle = true;
static bool endsDefined = false;

static void read_config()
{
    KConfig conf("kwinbeetoorc");
    conf.setGroup("General");
    colored_frame = conf.readBoolEntry("UseTitleBarBorderColors", false);
    do_draw_handle = conf.readBoolEntry("DrawGrabHandle", true);
    thickness = conf.readUnsignedNumEntry("FrameThickness", 5);
    tileName = conf.readEntry("TitleEffectName", "");
    QString opString = conf.readEntry("MenuButtonDoubleClickOperation", "NoOp");
    if (opString == "CloseWindow") {
        menu_dbl_click_op = 2;
    } else if (opString == "IconifyWindow") {
        menu_dbl_click_op = 1;
    } else {
        menu_dbl_click_op = 0;
    }
}

static void drawB2Rect(KPixmap *pix, const QColor &primary, bool down)
{
    QColor hColor = primary.light(150);
    QColor lColor = primary.dark(150);

    if (down) {
        // Swap colors
	QColor tColor = hColor;
	hColor = lColor;
	lColor = tColor;
    }
    
    if (QPixmap::defaultDepth() > 8) {
	KPixmapEffect::gradient(*pix, 
		hColor, lColor, KPixmapEffect::DiagonalGradient);
    } else {
        pix->fill(primary);
    }
    int x2 = pix->width() - 1;
    int y2 = pix->height() - 1;
    QPainter p(pix);
    p.setPen(lColor);
    p.drawLine(0, 0, x2, 0);
    p.drawLine(0, 0, 0, y2);
    p.drawLine(1, y2 - 1, x2 - 1, y2 - 1);
    p.drawLine(x2 - 1, 1, x2 - 1, y2 - 1);
    p.setPen(hColor);
    p.drawRect(1, 1, x2, y2);
}

static void create_pixmaps();
static void redraw_pixmaps();

QPixmap* kwin_get_menu_pix_hack()
{
    //return menu_pix;  FIXME
    return PIXMAP_A(P_MENU);
}

static void create_pixmaps()
{
    if ( pixmaps_created )
        return;
    pixmaps_created = true;

    int i;
    for (i = 0; i < NUM_PIXMAPS; i++) {
        pixmap[i] = new KPixmap;
	switch (i / NumStates) {
  	    case P_MAX : break;  // will be initialized by copying P_CLOSE
  	    case P_ICONIFY : pixmap[i]->resize(10,10); break;
  	    default : pixmap[i]->resize(16, 16); break;
	}
    }

    // there seems to be no way to load X bitmaps from data properly, so
    // we need to create new ones for each mask :P
    QBitmap pinupMask(16, 16, pinup_mask_bits, true);
    PIXMAP_A(P_PINUP)->setMask(pinupMask);
    PIXMAP_I(P_PINUP)->setMask(pinupMask);
    QBitmap pindownMask(16, 16, pindown_mask_bits, true);
    PIXMAP_AD(P_PINUP)->setMask(pindownMask);
    PIXMAP_ID(P_PINUP)->setMask(pindownMask);

    QBitmap menuMask(16, 16, menu_mask_bits, true);
    for (i = 0; i < NumStates; i++) 
	pixmap[P_MENU * NumStates + i]->setMask(menuMask);

    QBitmap helpMask(16, 16, help_mask_bits, true);
    for (i = 0; i < NumStates; i++) 
	pixmap[P_HELP * NumStates + i]->setMask(helpMask);

    QBitmap normalizeMask(16, 16, true);
    // draw normalize icon mask
    QPainter mask(&normalizeMask);
	
    QBrush one(Qt::color1);
    mask.fillRect(3, 3, 12, 12, one);
    mask.fillRect(0, 0, 10, 10, one);
    
    for (i = 0; i < NumStates; i++) 
	pixmap[P_NORMALIZE * NumStates + i]->setMask(normalizeMask);
    
    for (i = TitleActive; i < NumTitlePixmaps; i++) {
	titleGradient[i] = 0;
    }

    redraw_pixmaps();
}

static void delete_pixmaps()
{
    for (int i = 0; i < NUM_PIXMAPS; i++) {
    	delete pixmap[i];
	pixmap[i] = 0;
    }
    for (int i = TitleActive; i < NumTitlePixmaps; i++) {
    	delete titleGradient[i];
	titleGradient[i] = 0;
	
    }
    pixmaps_created = false;
}


B2Button::B2Button(Client *_client, QWidget *parent, const QString& tip, int pixIndex)
   : KWinButton(parent, 0, tip), hover(false)
{ 
    setBackgroundMode(NoBackground);
    client = _client;
    useMiniIcon = false;
    setFixedSize(16, 16);
    setPixmaps(pixIndex);
};

void B2Button::enterEvent(QEvent *e)
{
    KWinButton::enterEvent(e);
	
    hover = true;
    repaint(false);
}

void B2Button::leaveEvent(QEvent *e)
{
    KWinButton::leaveEvent(e);
	
    hover = false;
    repaint(false);
}

QSize B2Button::sizeHint() const
{
    return(QSize(16, 16));
}

QSizePolicy B2Button::sizePolicy() const
{
    return(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
}

void B2Button::drawButton(QPainter *p)
{
    QPixmap* gradient = titleGradient[client->isActive() ? 0 : 1];
    if (gradient) { 
	p->drawTiledPixmap(0, 0, width(), height(), *gradient, x() - 1, 1);
    } else { 
	p->fillRect(rect(), bg);
    }
    
    if (useMiniIcon && !client->miniIcon().isNull()) {
        QPixmap miniIcon = client->miniIcon();
        p->drawPixmap((width() - miniIcon.width()) / 2,
                      (height() - miniIcon.height()) / 2, miniIcon);
    } else {
	int type;
        if (client->isActive()) {
            if (isOn() || isDown())
		type = Down;
            else if (hover)
		type = Hover;
	    else
		type = Norm;
        } else {
            if (isOn() || isDown())
		type = IDown;
	    else if (hover)
		type = IHover;
            else
		type = INorm;
        }
	p->drawPixmap((width() - icon[type]->width()) / 2, 
		(height() - icon[type]->height()) / 2, *icon[type]);
    }
}

void B2Button::setPixmaps(int button_id)
{
    button_id *= NumStates;
    for (int i = 0; i < NumStates; i++) {
	icon[i] = B2::pixmap[button_id + i];
    }
    repaint(false);
}

void B2Button::mousePressEvent( QMouseEvent* e )
{
    last_button = e->button();
    QMouseEvent me ( e->type(), e->pos(), e->globalPos(), 
                     LeftButton, e->state() );
    KWinButton::mousePressEvent( &me );
}

void B2Button::mouseReleaseEvent( QMouseEvent* e )
{
    QMouseEvent me ( e->type(), e->pos(), e->globalPos(), 
                     LeftButton, e->state() );
    KWinButton::mouseReleaseEvent( &me );
}

B2Titlebar::B2Titlebar(B2Client *parent)
    : QWidget(parent, 0, WStyle_Customize | WRepaintNoErase),
      set_x11mask(false), isfullyobscured(false), shift_move(false),
      client(parent)
{
    setBackgroundMode(NoBackground);
    captionSpacer = new QSpacerItem(10, 20, 
	    QSizePolicy::Expanding, QSizePolicy::Fixed);
    endLabel[0] = new QSpacerItem(endWidth[0], 20, 
	    QSizePolicy::Fixed, QSizePolicy::Fixed);
    endLabel[1] = new QSpacerItem(endWidth[1], 20, 
	    QSizePolicy::Fixed, QSizePolicy::Fixed);
}

bool B2Titlebar::x11Event(XEvent *e)
{
    if (!set_x11mask) {
	set_x11mask = true;
	XSelectInput(qt_xdisplay(), winId(),
	    KeyPressMask | KeyReleaseMask |
	    ButtonPressMask | ButtonReleaseMask |
	    KeymapStateMask |
	    ButtonMotionMask |
	    EnterWindowMask | LeaveWindowMask |
	    FocusChangeMask |
	    ExposureMask |
	    PropertyChangeMask |
	    StructureNotifyMask | SubstructureRedirectMask |
	    VisibilityChangeMask);
    }
    switch( e->type ) {
    case VisibilityNotify:
	isfullyobscured = false;
	if (e->xvisibility.state == VisibilityFullyObscured) {
	    isfullyobscured = true;
	    client->unobscureTitlebar();
	}
	break;
    default:
	break;
    }
    return QWidget::x11Event(e);
}

void B2Titlebar::drawTitleBar(QPainter &p, QRect &t, bool state)
{
    int activeState = state ? 0 : 1;
    QPixmap* gradient = titleGradient[activeState];
    QPixmap* leftGradient = titleGradient[LeftActive + activeState];
    QPixmap* rightGradient = titleGradient[RightActive + activeState];
    
    // black titlebar frame
    p.setPen(Qt::black);
    p.drawRect(t);
    // titlebar fill
    if (gradient) {
	if (endsDefined) {
	    p.drawPixmap(1, 1, *leftGradient);
	    p.drawTiledPixmap(1 + endWidth[0], 1, 
		    t.right() - 1 - endWidth[0] - endWidth[1], t.height() - 2, 
		    *gradient);
	    p.drawPixmap(t.right() - endWidth[1], 1, *rightGradient);
	} else {
	    p.drawTiledPixmap(1, 1, t.right() - 1, t.height() - 2, *gradient);
	}
    } else {
	qDrawShadeRect(&p, 1, 1, t.right() - 1, t.height() - 1, 
		options->colorGroup(Options::TitleBar, state), 
		false, 1, 0, 
		&options->colorGroup(Options::TitleBar, state).brush(
		    QColorGroup::Button));
    }
    // and the caption
    p.setPen(options->color(Options::Font, state));
    p.setFont(options->font(state));
    t = captionSpacer->geometry();
    p.drawText(t, AlignLeft | AlignVCenter, client->caption());
}

void B2Titlebar::recalcBuffer()
{
    QFontMetrics fm(options->font(true));
    titleBuffer.resize(width(), height());

    QPainter p(&titleBuffer);
    QRect t = rect();

    drawTitleBar(p, t, true);
    
    oldTitle = caption();
}

void B2Titlebar::resizeEvent(QResizeEvent *) 
{
    recalcBuffer();
    repaint(false);
}

void B2Titlebar::paintEvent(QPaintEvent *)
{
    if (client->isActive())
        bitBlt(this, 0, 0, &titleBuffer, 0, 0, 
		titleBuffer.width(), titleBuffer.height(), Qt::CopyROP, true);
    else {
        QPainter p(this);
        QRect t = rect();

	drawTitleBar(p, t, false);
    }
}

void B2Titlebar::mouseDoubleClickEvent( QMouseEvent * )
{
    client->workspace()->performWindowOperation( client, 
	    options->operationTitlebarDblClick() );
}

void B2Titlebar::mousePressEvent( QMouseEvent * e )
{
    shift_move = e->state() & ShiftButton;
    if (shift_move) {
        moveOffset = e->globalPos();
    }
    QMouseEvent _e(QEvent::MouseButtonPress, mapToParent(e->pos()),
        e->globalPos(), e->button(), e->state());
    client->mousePressEvent( &_e);
}

void B2Titlebar::mouseReleaseEvent( QMouseEvent * e )
{
    shift_move = false;
    QMouseEvent _e(QEvent::MouseButtonRelease, mapToParent(e->pos()),
        e->globalPos(), e->button(), e->state());
    client->mouseReleaseEvent( &_e);
}

void B2Titlebar::mouseMoveEvent( QMouseEvent * e )
{
    if (shift_move) {
	int oldx = mapFromGlobal(moveOffset).x();
        int xdiff = e->globalPos().x() - moveOffset.x();
        moveOffset = e->globalPos();
	if (oldx >= 0 && oldx <= rect().right()) {
            client->titleMoveRel(xdiff);
	}
    } else {
	QMouseEvent _e(QEvent::MouseMove, mapToParent(e->pos()),
	    e->globalPos(), e->button(), e->state());
	client->mouseMoveEvent( &_e);
    }
}

void B2Client::maxButtonClicked( )
{
    switch ( button[BtnMax]->last_button ) {
    case MidButton:
	maximize( MaximizeVertical );
	break;
    case RightButton:
	maximize( MaximizeHorizontal );
	break;
    default: //LeftButton:
	maximize( MaximizeFull );
	break;
    }
}

B2Client::B2Client( Workspace *ws, WId w, QWidget *parent,
                            const char *name )
    : Client( ws, w, parent, name, WResizeNoErase | WRepaintNoErase ),
      bar_x_ofs(0), in_unobs(0)
{
    const QString tips[] = {
	i18n("Menu"), i18n("Sticky"), 
        i18n("Minimize"), i18n("Maximize"),
        i18n("Close"), i18n("Help") 
    };

    draw_handle = do_draw_handle;
    setBackgroundMode(NoBackground);

    // Set button pointers to NULL so we know what has been created
    for (int i = 0; i < BtnCount; i++)
        button[i] = NULL;        

    g = new QGridLayout(this, 0, 0);
    g->addMultiCellWidget(windowWrapper(), 1, 1, 1, 2);

    // Bottom border height
    spacer = new QSpacerItem(16, thickness + (mustDrawHandle() ? 4 : 0), 
	    QSizePolicy::Expanding, QSizePolicy::Minimum);
    // Left and right border width
    vspacer1 = new QSpacerItem(thickness, 16, 
	    QSizePolicy::Minimum, QSizePolicy::Expanding);
    vspacer2 = new QSpacerItem(thickness, 16, 
	    QSizePolicy::Minimum, QSizePolicy::Expanding);
    
    g->addItem(vspacer1, 1, 0);
    g->addColSpacing(1, 16);
    g->setColStretch(2, 1);
    g->setRowStretch(1, 1);
    g->addItem(vspacer2, 1, 3);
    g->addItem(spacer, 2, 1);
    // titlebar
    g->addRowSpacing(0, 20);

    titlebar = new B2Titlebar(this);
    titlebar->setMinimumWidth(16);
    titlebar->setFixedHeight(20);

    QBoxLayout *titleLayout = new QBoxLayout(titlebar, 
	    QBoxLayout::LeftToRight, 0, 1, 0);
    titleLayout->addSpacing(1);
    titleLayout->addItem(titlebar->endLabel[0]);

    if (options->customButtonPositions()) {
        addButtons(options->titleButtonsLeft(), tips, titlebar, titleLayout);
        titleLayout->addItem(titlebar->captionSpacer);
        addButtons(options->titleButtonsRight(), tips, titlebar, titleLayout);
    } else {
        addButtons("MSH", tips, titlebar, titleLayout);
        titleLayout->addItem(titlebar->captionSpacer);
        addButtons("IAX", tips, titlebar, titleLayout);
    }

    titleLayout->addItem(titlebar->endLabel[1]);
    titleLayout->addSpacing(1);
 
    QColor c = options->colorGroup(Options::TitleBar, isActive()).
        color(QColorGroup::Button);

    for (int i = 0; i < BtnCount; i++)
        if (button[i])
            button[i]->setBg(c);

    titlebar->recalcBuffer();
    positionButtons();

    connect(options, SIGNAL(resetClients()), this, SLOT(slotReset()));
}

void B2Client::addButtons(const QString& s, const QString tips[],
                          B2Titlebar* tb, QBoxLayout* titleLayout)
{
    if (s.length() <= 0)
	return;

    for (unsigned int i = 0; i < s.length(); i++) {
        switch (s[i].latin1()) {
            case 'M':  // Menu button
                if (!button[BtnMenu]) {
                    button[BtnMenu] = new B2Button(this, 
			    tb, tips[BtnMenu], P_MENU);
                    button[BtnMenu]->setUseMiniIcon();
                    connect(button[BtnMenu], SIGNAL(clicked()),
                            this, SLOT(menuButtonPressed()));
                    titleLayout->addWidget(button[BtnMenu]);
                }
                break;
            case 'S':  // Sticky button
                if (!button[BtnSticky]) {
                    button[BtnSticky] = new B2Button(this, 
			    tb, tips[BtnSticky], P_PINUP);
                    button[BtnSticky]->setToggle();
                    button[BtnSticky]->setDown(isSticky());
                    connect(button[BtnSticky], SIGNAL(clicked()),
                            this, SLOT(toggleSticky()));
                    titleLayout->addWidget(button[BtnSticky]);
                }
                break;
            case 'H':  // Help button
                if (providesContextHelp() && (!button[BtnHelp])) {
                    button[BtnHelp] = new B2Button(this, 
			    tb, tips[BtnHelp], P_HELP);
                    connect(button[BtnHelp], SIGNAL(clicked()),
                            this, SLOT(contextHelp()));
                    titleLayout->addWidget(button[BtnHelp]);
                }
                break;
            case 'I':  // Minimize button
                if (isMinimizable() && (!button[BtnIconify])) {
                    button[BtnIconify]= new B2Button(this, 
			    tb, tips[BtnIconify], P_ICONIFY);
                    connect(button[BtnIconify], SIGNAL(clicked()),
                            this, SLOT(iconify()));
                    titleLayout->addWidget(button[BtnIconify]);
                }
                break;
            case 'A':  // Maximize button
                if (isMaximizable() && (!button[BtnMax])) {
                    button[BtnMax]= new B2Button(this, 
			    tb, tips[BtnMax], isMaximized() ? P_NORMALIZE : P_MAX);
                    connect(button[BtnMax], SIGNAL(clicked()),
                            this, SLOT(maxButtonClicked()));
                    titleLayout->addWidget(button[BtnMax]);
                }
                break;
            case 'X':  // Close button
                if (isCloseable() && !button[BtnClose]) {
                    button[BtnClose]= new B2Button(this, tb, tips[BtnClose], P_CLOSE);
                    connect(button[BtnClose], SIGNAL(clicked()),
                            this, SLOT(closeWindow()));
                    titleLayout->addWidget(button[BtnClose]);
                }
		break;
	    case '_': // Additional spacing
		titleLayout->addSpacing(4);
		break;
	}
    } 
}


void B2Client::iconChange()
{
    // TODO Build an icon set with the hilighted and disabled icons.
    if (button[BtnMenu])
        button[BtnMenu]->repaint(false);   
}


// Gallium: New button show/hide magic for customizable
//          button positions.
void B2Client::calcHiddenButtons()
{
    // Hide buttons in this order:
    // Sticky, Help, Maximize, Minimize, Close, Menu
    B2Button* btnArray[] = { button[BtnSticky], button[BtnHelp],
                             button[BtnMax], button[BtnIconify],
                             button[BtnClose], button[BtnMenu] };
    int minWidth = 120; 
    int currentWidth = width();
    int count = 0;
    int i;

    // Determine how many buttons we need to hide
    while(currentWidth < minWidth) {
        currentWidth += 17;  // Allow for spacer (extra 1pix)
        count++;
    }
    // Bound the number of buttons to hide
    if (count > BtnCount) count = BtnCount;

    // Hide the required buttons
    for(i = 0; i < count; i++) {
        if (btnArray[i] && btnArray[i]->isVisible())
            btnArray[i]->hide();
    }
    // Show the rest of the buttons
    for(i = count; i < BtnCount; i++) {
        if (btnArray[i] && (!btnArray[i]->isVisible()))
            btnArray[i]->show();
    }
}

void B2Client::resizeEvent( QResizeEvent* e)
{
    Client::resizeEvent( e );
    calcHiddenButtons();
    titlebar->layout()->activate();
    positionButtons();

    /* may be the resize cuted off some space occupied by titlebar, which
       was moved, so instead of reducing it, we first try to move it */
    titleMoveAbs(bar_x_ofs);
    doShape();

    repaint(); //there is some strange wrong repaint of the frame without
}

void B2Client::captionChange( const QString &)
{
    positionButtons();
    titleMoveAbs(bar_x_ofs);
    doShape();
    titlebar->recalcBuffer();
    titlebar->repaint(false);
}

void B2Client::paintEvent( QPaintEvent* e)
{
    // If the window is shaded, there frame must not be drawn.
    if (isShade()) return;
    
    QPainter p(this);
    p.setPen(Qt::black);

    QRect t = titlebar->geometry();
    
    // Frame height, this is used a lot of times
    int fHeight = height() - t.height();

    // distance from the bottom border - it is different if window is resizable
    int bb = mustDrawHandle() ? 4 : 0;

    Options::ColorType frameColor = colored_frame ? Options::TitleBar : Options::Frame;
    QColorGroup fillColor = options->colorGroup(frameColor, isActive());
    QBrush fillBrush(options->color(frameColor, isActive()));

    // outer frame rect
    p.drawRect(0, t.bottom() - thickness + 1, width(), fHeight - bb + thickness);

    if (thickness >= 2) {
	// inner window rect
	p.drawRect(thickness - 1, t.bottom(), width() - 2 * (thickness - 1), 
		fHeight - thickness + 2 - bb);
	
	if (thickness >= 3) {
	    // frame shade panel
	    qDrawShadePanel(&p, 1, t.bottom() - thickness + 2, 
	    	width() - 2, fHeight - 2 - bb + thickness, fillColor, false);
	    if (thickness >= 4) {
		qDrawShadePanel(&p, thickness - 2, 
	    	    t.bottom() - 1, width() - 2 * (thickness - 2), 
	    	    fHeight + 4 - thickness - bb, fillColor, true);
		if (thickness >= 5) {
		    // draw frame interior
		    p.fillRect(2, t.bottom() - thickness + 3, 
		    	width() - 4, thickness - 4, fillBrush);
		    p.fillRect(2, height() - thickness - bb + 2, 
		    	width() - 4, thickness - 4, fillBrush);
		    p.fillRect(2, t.bottom() - 1, 
		    	thickness - 4, fHeight - thickness - bb  + 4, fillBrush);
		    p.fillRect(width() - thickness + 2, t.bottom() - 1,
		    	thickness - 4, fHeight - thickness - bb + 4, fillBrush);
		}
	    }
	}
    }
    
    //bottom handle rect
    if (mustDrawHandle()) {
        p.setPen(Qt::black);
	int hx = width() - 40;
	int hw = 40;

	p.drawLine(width() - 1, height() - thickness - 4, 
		width() - 1, height() - 1);
	p.drawLine(hx, height() - 1, width() - 1, height() - 1);
	p.drawLine(hx, height() - 4, hx, height() - 1);

	p.fillRect(hx + 1, height() - thickness - 3, 
		hw - 2, thickness + 2, fillBrush);

	p.setPen(fillColor.dark()); 
	p.drawLine(width() - 2, height() - thickness - 4, 
		width() - 2, height() - 2);
	p.drawLine(hx + 1, height() - 2, width() - 2, height() - 2);

	p.setPen(fillColor.light());
	p.drawLine(hx + 1, height() - thickness - 2, 
		hx + 1, height() - 3);
	p.drawLine(hx + 1, height() - thickness - 3, 
		width() - 3, height() - thickness - 3);
    } 

    /* OK, we got a paint event, which means parts of us are now visible
       which were not before. We try the titlebar if it is currently fully
       obscured, and if yes, try to unobscure it, in the hope that some
       of the parts which we just painted were in the titlebar area.
       It can happen, that the titlebar, as it got the FullyObscured event
       had no chance of becoming partly visible. The problem is, that
       we now might have the space available, but the titlebar gets no
       visibilitinotify events until its state changes, so we just try
     */
    if (titlebar->isFullyObscured()) {
        /* We first see, if our repaint contained the titlebar area */
	QRegion reg(QRect(0, 0, width(), 20));
	reg = reg.intersect(e->region());
	if (!reg.isEmpty())
	    unobscureTitlebar();
    }
}

#define QCOORDARRLEN(x) sizeof(x)/(sizeof(QCOORD)*2)

void B2Client::doShape()
{
    QRect t = titlebar->geometry();
    QRegion mask(rect());
    
    // Calculate mask if the window is not shaded only
    if (isShade()) {
	mask = t;
    } else {
	// top to the tilebar right
	if (bar_x_ofs) {
	    mask -= QRect(0, 0, 
		    bar_x_ofs, t.height() - thickness); //left from bar
	    mask -= QRect(0, t.height() - thickness, 
		    1, 1); //top left point
	}
	if (t.right() < width() - 1) {
	    mask -= QRect(width() - 1, t.height() - thickness, 
		    1, 1); // top right point
	    mask -= QRect(t.right() + 1, 0, 
		    width() - t.right() - 1, t.height() - thickness);
	}
	mask -= QRect(width() - 1, height() - 1, 1, 1); // bottom right point
	if (mustDrawHandle()) {
	    mask -= QRect(0, height() - 5, 1, 1); // bottom left point
	    mask -= QRect(width() - 1, height() - 1, 1, 1); // bottom right point
	    mask -= QRect(width() - 40, height() - 1, 1, 1); // handle left point
	    mask -= QRect(0, height() - 4, width() - 40, 4); // bottom left
	} else {
	    mask -= QRect(0, height() - 1, 1, 1); // bottom left point
	}
    }
    setMask(mask);
}

void B2Client::showEvent(QShowEvent *ev)
{
    calcHiddenButtons();
    Client::showEvent(ev);
    doShape();
    repaint();
    titlebar->repaint(false);
}

void B2Client::windowWrapperShowEvent( QShowEvent* )
{
    doShape();
}

Client::MousePosition B2Client::mousePosition( const QPoint& p ) const
{
    const int range = 16;
    QRect t = titlebar->geometry();
    t.setHeight(20 - thickness);
    int ly = t.bottom();
    int lx = t.right();
    int bb = mustDrawHandle() ? 0 : 5;

    if (p.x() > t.right()) {
        if (p.y() <= ly + range && p.x() >= width() - range)
            return TopRight;
        else if ( p.y() <= ly + thickness)
            return Top;
    } else if (p.x() < bar_x_ofs) {
        if (p.y() <= ly + range && p.x() <= range)
            return TopLeft;
        else if (p.y() <= ly + thickness)
            return Top;
    } else if (p.y() < ly) {
        if (p.x() > bar_x_ofs+thickness && p.x() < lx - thickness && p.y() > thickness)
            return Client::mousePosition(p);
	else if (p.x() > bar_x_ofs+range && p.x() < lx - range)
            return Top;
	else if (p.y() <= range) {
            if (p.x() <= bar_x_ofs + range) return TopLeft;
            else return TopRight;
        } else {
            if (p.x() <= bar_x_ofs + range) return Left;
            else return Right;
        }
    }

    if (p.y() >= height() - 8 + bb) {
        /* the normal Client:: only wants border of 4 pixels */
	if (p.x() <= range) return BottomLeft;
	if (p.x() >= width() - range) return BottomRight;
	return Bottom;
    }

    return Client::mousePosition(p);
}


void B2Client::titleMoveAbs(int new_ofs)
{
    if (new_ofs < 0) new_ofs = 0;
    if (new_ofs + titlebar->width() > width()) {
        new_ofs = width() - titlebar->width();
    }
    if (bar_x_ofs != new_ofs) {
        bar_x_ofs = new_ofs;
	positionButtons();
	doShape();
	repaint( 0, 0, width(), 20, false );
	titlebar->repaint(false);
    }
}

void B2Client::titleMoveRel(int xdiff)
{
    titleMoveAbs(bar_x_ofs + xdiff);
}

void B2Client::stickyChange(bool on)
{
    if (button[BtnSticky]) {
        button[BtnSticky]->setDown(on);
        button[BtnSticky]->setTipText(on ? i18n("Un-Sticky") : i18n("Sticky"));
    }
}

void B2Client::maximizeChange(bool m)
{
    if (button[BtnMax]) {
        button[BtnMax]->setPixmaps( m ? P_NORMALIZE : P_MAX );
        button[BtnMax]->repaint();
        button[BtnMax]->setTipText(m ? i18n("Restore") : i18n("Maximize"));
    }
    spacer->changeSize(10, thickness + (mustDrawHandle() ? 4 : 0), 
	    QSizePolicy::Expanding, QSizePolicy::Minimum);

    g->activate();
    doShape();
    repaint(false);
}

void B2Client::activeChange(bool on)
{
    repaint(false);
    titlebar->repaint(false);

    QColor c = options->colorGroup(Options::TitleBar, on).
        color(QColorGroup::Button);

    for(int i = 0; i < BtnCount; i++)
        if (button[i]) {
           button[i]->setBg(c);
           button[i]->repaint(false);
        }
}

void B2Client::menuButtonPressed()
{
    static B2Client *lastClient = NULL;
    
    bool dbl = (lastClient == this && 
	        time.elapsed() <= QApplication::doubleClickInterval());
    lastClient = this;
    time.start();
    if (!dbl) {
	QPoint menupoint = button[BtnMenu]->mapToGlobal(button[BtnMenu]->
				  rect().bottomLeft());
	workspace()->showWindowMenu( menupoint, this );
	button[BtnMenu]->setDown(false);
    } else if (menu_dbl_click_op == 2) {
          closeWindow();
    } else if (menu_dbl_click_op == 1) {
	iconify();
    }
}

void B2Client::slotReset()
{
    redraw_pixmaps();
    QColor c = options->colorGroup(Options::TitleBar, isActive()).
        color(QColorGroup::Button);

    for (int i = 0; i < BtnCount; i++)
        if (button[i]) {
            button[i]->setBg(c);
            button[i]->repaint(false);
        }

    repaint();
    titlebar->recalcBuffer();
    titlebar->repaint(false);
    titlebar->endLabel[0]->changeSize(endWidth[0], 18,
	    QSizePolicy::Fixed, QSizePolicy::Fixed);
    titlebar->endLabel[1]->changeSize(endWidth[1], 18,
	    QSizePolicy::Fixed, QSizePolicy::Fixed);
}

void B2Client::unobscureTitlebar()
{
    /* we just noticed, that we got obscured by other windows
       so we look at all windows above us (stacking_order) merging their
       masks, intersecting it with our titlebar area, and see if we can
       find a place not covered by any window */
    if (in_unobs) {
	return;
    }
    in_unobs = 1;
    QRegion reg(QRect(0,0,width(),20));
    ClientList::ConstIterator it = workspace()->stackingOrder().find(this);
    ++it;
    while (it != workspace()->stackingOrder().end()) {
        /* the clients all have their mask-regions in local coords
	   so we have to translate them to a shared coord system
	   we choose ours */
	int dx = (*it)->x() - x();
	int dy = (*it)->y() - y();
	QRegion creg = (*it)->getMask();
	creg.translate(dx, dy);
	reg -= creg;
	if (reg.isEmpty()) {
	    // early out, we are completely obscured
	    break;
	}
	++it;
    }
    if (!reg.isEmpty()) {
        // there is at least _one_ pixel from our title area, which is not
	// obscured, we use the first rect we find
	// for a first test, we use boundingRect(), later we may refine
	// to rect(), and search for the nearest, or biggest, or smthg.
	titleMoveAbs(reg.boundingRect().x());
    }
    in_unobs = 0;
}

static void redraw_pixmaps()
{
    QColorGroup aGrp = options->colorGroup(Options::ButtonBg, true);
    QColorGroup iGrp = options->colorGroup(Options::ButtonBg, false);

    // close
    drawB2Rect(PIXMAP_A(P_CLOSE), aGrp.button(), false);
    drawB2Rect(PIXMAP_AH(P_CLOSE), aGrp.button(), true);
    drawB2Rect(PIXMAP_AD(P_CLOSE), aGrp.button(), true);

    drawB2Rect(PIXMAP_I(P_CLOSE), iGrp.button(), false);
    drawB2Rect(PIXMAP_IH(P_CLOSE), iGrp.button(), true);
    drawB2Rect(PIXMAP_ID(P_CLOSE), iGrp.button(), true);

    // maximize
    int i;
    for (i = 0; i < NumStates; i++) {
	*pixmap[P_MAX * NumStates + i] = *pixmap[P_CLOSE * NumStates + i];
	pixmap[P_MAX * NumStates + i]->detach();
    }

    // normalize + iconify
    KPixmap smallBox;
    smallBox.resize(10, 10);
    KPixmap largeBox;
    largeBox.resize(12, 12);

    for (i = 0; i < NumStates; i++) {
	bool is_act = (i < 3);
	bool is_down = (i == Down || i == IDown);
	KPixmap *pix = pixmap[P_NORMALIZE * NumStates + i];
	drawB2Rect(&smallBox, is_act ? aGrp.button() : iGrp.button(), is_down);
	drawB2Rect(&largeBox, is_act ? aGrp.button() : iGrp.button(), is_down);
	pix->fill(options->color(Options::TitleBar, is_act));
	bitBlt(pix, 3, 3, &largeBox, 0, 0, 12, 12, Qt::CopyROP, true);
	bitBlt(pix, 0, 0, &smallBox, 0, 0, 10, 10, Qt::CopyROP, true);

	bitBlt(pixmap[P_ICONIFY * NumStates + i],
	       0, 0, &smallBox, 0, 0, 10, 10, Qt::CopyROP, true);
    }

    QPainter p;
    // x for close + menu + help
    for (int j = 0; j < 3; j++) {
        int pix;
        unsigned const char *light, *dark;
        switch (j) {
        case 0 :
            pix = P_CLOSE; light = close_white_bits; dark = close_dgray_bits;
            break;
        case 1 :
            pix = P_MENU; light = menu_white_bits; dark = menu_dgray_bits;
            break;
        default:
            pix = P_HELP; light = help_light_bits; dark = help_dark_bits;
            break;
        }
        for (i = 0; i < NumStates; i++) {
            p.begin(pixmap[pix * NumStates + i]);
            kColorBitmaps(&p, (i < 3) ? aGrp : iGrp, 0, 0, 16, 16, true,
                          light, NULL, NULL, dark, NULL, NULL);
            p.end();
        }
    }

    // pin
    for (i = 0; i < NumStates; i++) {
	bool isDown = (i == Down || i == IDown);
        unsigned const char *white = isDown ? pindown_white_bits : pinup_white_bits;
        unsigned const char *gray = isDown ? pindown_gray_bits : pinup_gray_bits;
        unsigned const char *dgray =isDown ? pindown_dgray_bits : pinup_dgray_bits;
        p.begin(pixmap[P_PINUP * NumStates + i]);
        kColorBitmaps(&p, (i < 3) ? aGrp : iGrp, 0, 0, 16, 16, true, white,
                      gray, NULL, dgray, NULL, NULL);
        p.end();
    }

    // Apply the hilight effect to the 'Hover' icons
    KIconEffect ie;
    QPixmap hilighted;
    for (i = 0; i < P_NUM_BUTTON_TYPES; i++) {
	int offset = i * NumStates;
	hilighted = ie.apply(*pixmap[offset + Norm], 
		KIcon::Small, KIcon::ActiveState);
	*pixmap[offset + Hover] = hilighted;    

	hilighted = ie.apply(*pixmap[offset + INorm], 
		KIcon::Small, KIcon::ActiveState);
	*pixmap[offset + IHover] = hilighted;    
    }

    
    // Create the titlebar gradients
    QImage titleTheme(tileName);
    QString b2Segments = titleTheme.text("b2segments");
    if (!b2Segments.isEmpty()) {
	endsDefined = true;
    	endWidth[0] = b2Segments.section(',', 0, 0).toInt();
    	endWidth[1] = b2Segments.section(',', 1, 1).toInt();
    } else {
	endsDefined = false;
	endWidth[0] = endWidth[1] = 3;
    }
    
    if (!titleTheme.isNull() && QPixmap::defaultDepth() > 8) {
	QColor titleColor[4] = {
	    options->color(Options::TitleBlend, true),
            options->color(Options::TitleBar, true),

	    options->color(Options::TitleBlend, false),
	    options->color(Options::TitleBar, false)
	};

	for (i = 0; i < 2; i++) {
	    if (titleColor[2 * i] != titleColor[2 * i + 1]) {
		QImage img = titleTheme;
		KImageEffect::flatten(img, 
			titleColor[2 * i], titleColor[2 * i + 1], 32);
		int w, xoff;
		w = endsDefined ? endWidth[0] + endWidth[1] : 0;
		xoff = endsDefined ? endWidth[0] : 0;
		if (!titleGradient[i]) {

		    titleGradient[i] = new QPixmap(
			    titleTheme.width() - w, 18);
		}
		p.begin(titleGradient[i]);
		p.drawImage(0, 0, img, xoff, 0);
		p.end();
		if (endsDefined) {
		    if (!titleGradient[LeftActive + i]) {
			titleGradient[LeftActive + i] = new QPixmap(
				endWidth[0], 18);
		    }
		    if (!titleGradient[RightActive + i]) {
			titleGradient[RightActive + i] = new QPixmap(
				endWidth[1], 18);
		    } 		
		    p.begin(titleGradient[LeftActive + i]);
		    p.drawImage(0, 0, img);
		    p.end();
		    p.begin(titleGradient[RightActive + i]);
		    p.drawImage(0, 0, img, img.width() - endWidth[1], 0);
		    p.end();
	        }
	    } else { 
	       delete titleGradient[i]; 
	       delete titleGradient[LeftActive + i];
	       delete titleGradient[RightActive + i];
	       titleGradient[i] = 0;
	       titleGradient[LeftActive + i] = 0;
	       titleGradient[RightActive + i] = 0;
	    }
	}
    }
}

void B2Client::positionButtons()
{
    QFontMetrics fm(options->font(isActive()));
    QString cap = caption();
    if ( cap.length() < 4 ) // make sure the titlebar has sufficiently wide
        cap = "XXXX";      // area for dragging the window
    int textLen = fm.width( cap );

    QRect t = titlebar->captionSpacer->geometry();
    int titleWidth = titlebar->width() - t.width() + textLen + 2;
    if (titleWidth > width()) titleWidth = width();

    titlebar->resize(titleWidth, 20);
    titlebar->move(bar_x_ofs, 0);
}

// Transparent bound stuff.

static QRect *visible_bound;

static QPointArray bound_shape;

void B2Client::drawbound(const QRect& geom)
{
    if (!visible_bound) {
	visible_bound = new QRect(geom);
	QRect t = titlebar->geometry();
	int frameTop = geom.top() + t.bottom();
	int barLeft = geom.left() + bar_x_ofs;
	int barRight = barLeft + t.width() - 1; 
	if (barRight > geom.right()) barRight = geom.right();
	
	bound_shape.putPoints(0, 8,
		geom.left(), frameTop,
		barLeft, frameTop,
		barLeft, geom.top(),
		barRight, geom.top(),
		barRight, frameTop,
		geom.right(), frameTop,
		geom.right(), geom.bottom(),
		geom.left(), geom.bottom());
    } else {
	*visible_bound = geom;
    }
    QPainter p(workspace()->desktopWidget());
    p.setPen(QPen(Qt::white, 5));
    p.setRasterOp(Qt::XorROP);
    p.drawPolygon(bound_shape);
}

void B2Client::clearbound()
{
    if (!visible_bound)
	return;
    drawbound(*visible_bound);
    delete visible_bound;
    visible_bound = 0;
}


};

extern "C"
{
    Client *allocate(Workspace *ws, WId w, int)
    {
       return new B2::B2Client(ws, w);
    }
    
    void init()
    {
       B2::read_config();
       B2::create_pixmaps();
    }
    
    void reset()
    {
       B2::read_config();
       B2::redraw_pixmaps();
       // Ensure change in tooltip state gets applied
       KWinInternal::Workspace::self()->slotResetAllClientsDelayed();
    }
    
    void deinit()
    {
       B2::delete_pixmaps();
    }
}

#include "beetooclient.moc"
/* vim: set sw=4 ts=8 : */
